
/* 
	Diese kombinierte Anwendung kann sowohl Bilder berblenden als
	auch mit dem ChromaKey-Verfahren kombinieren. Zwischen den beiden Betriebsarten 
	wird durch einen Kommandozeilenparameter (b oder c) umgeschaltet

	In beiden F"allen werden als Eingabe zwei 400x400 Bilder mit 8b je Farbkomponente
	 (Rot, Grn, Blau) erwartet. Das Ergebnisbild (im gleichen Format)
    wir dann in eine weitere PPM-Datei geschrieben. 

	Betriebsart "Uberblenden:
    =========================
	
    Der Mischfaktor alpha gibt hier an, ob das Ausgabebild nur aus den
    Pixeln des ersten Bildes (alpha=255), nur denen des 2. Bildes
    (alpha=0), oder einer gewichteten Summe bestehen soll (z.B. bei
    alpha=128 gehen die beiden Pixel hlftig ein.)
    
   Aufruf:

      ./chromablend image1.ppm image2.ppm b 255 foo.ppm

      image1.ppm wird voll gewichtet in foo.ppm geschrieben

      ./chromablend image1.ppm image2.ppm b 128 foo.ppm

      image1.ppm und image2.ppm werden berblendet in foo.ppm geschrieben


		Der Chroma-Key ist dabei eine 3x8b RGB-Farbe. Alle Pixel aus dem 1.
		Bild bei denen die Summe der komponentenweisen Abweichungen vom Chroma-Key
		nur +/- epsilon betragen, werden in der Ausgabe durch
		die entsprechenden Pixel des 2.  Bildes ersetzt. Alle anderen
		werden direkt in die Ausgabe kopiert.

    Betriebsart Chroma-Key:
    =======================
                             
   Aufruf:

      ./chromablend image1.ppm image2.ppm c 255 5 foo.ppm

			Alle Pixel aus image1.ppm die blau sind (komponentenweise insgesamt nur um 5 von
			der Farbe R=0, G=0, B=255 abweichen)
		  werden durch die Pixel aus image2.ppm ersetzt
			und in foo.ppm geschrieben. Alle anderen Pixel werden aus
			image1.ppm einfach kopiert.
      
*/

#include <stdio.h>
   
#define HEIGHT 300
#define WIDTH  400
#define NUMPIXELS (HEIGHT*WIDTH)
#define COMPONENTS (NUMPIXELS*3)  // 3 farbkanle (rot,grn,blau) je Pixel
#define IMGSIZE (COMPONENTS*sizeof(PIXEL))
typedef unsigned char PIXEL, ALPHA;


/* =======Hier geht der spannende Teil los========================== */

/* Bilder berblenden */

PIXEL *blend(PIXEL image1[], PIXEL image2[], ALPHA alpha, PIXEL oimage[]) {

  unsigned int n;
  
    // Ganzes Bild durchlaufen (die RGB-Komponenten aller Pixel)

    for (n=0; n<IMGSIZE; ++n) {
      // grerer zahlenbereich (fixpunkt, alle Werte *255)
      unsigned short temp;

      // Bildpunkte mischen
      temp = alpha * image1[n] + (255-alpha) * image2[n];

      // Jetzt Komponente wieder rckwandeln in bereich 0..255
      oimage[n] = temp >> 8;
    } 

}

/* Bilder mit Chroma-Key-Verfahren mischen */

// kleines Makro fr Betragsfunktion
#define ABS(x) (((x) >= 0) ? (x) : -(x))

void
chromakey(
	PIXEL image1[], // RGB-Komponenten des 1. Bildes mit Key Color
	PIXEL image2[], // RGB-Komponenten des 2. Bildes (eingeblendet)
	PIXEL oimage[], // RGB-Komponenten des Augabebildes
	unsigned int chromakey, // 24b RGB-Wert der Key Color
	unsigned int epsilon	// Erlaubte Abweichung von der Key Color
	) {

	// 24b Chroma-Key-Color in drei 8b Komponenten (RGB) zerlegen
	unsigned char r = (chromakey >> 16), g = (chromakey >> 8), b = chromakey;
	unsigned int n, delta;
  
	// Ganzes Bild durchlaufen (die RGB-Komponenten aller Pixel)
	for (n=0; n<IMGSIZE; n+=3) {

		delta = ABS(image1[n+0]-r) + ABS(image1[n+1]-g) + ABS(image1[n+2]-b);
	
		if (delta	<= epsilon) {  // Chroma-Key passt, Pixel ersetzen

			oimage[n+0] = image2[n+0];
			oimage[n+1] = image2[n+1];
			oimage[n+2] = image2[n+2];

		} else { // Chroma-Key passt nicht, alte Pixel "ubernehmen

			oimage[n+0] = image1[n+0];
			oimage[n+1] = image1[n+1];
			oimage[n+2] = image1[n+2];

		}
	}
}


/* =====Ab hier wird es wieder langweilig=========================== */

/* einlesen einer passenden ppm datei */

PIXEL *readimage(char *filename) {
  FILE *infile = fopen(filename, "r");
  PIXEL *image = (PIXEL*) malloc(IMGSIZE);
  char buf[256];

  if (infile && image) {
    fgets(buf, 80, infile);   // skip P6
    fgets(buf, 80, infile);   // skip # CREATOR
    fgets(buf, 80, infile);   // skip dimensions
    fgets(buf, 80, infile);   // skip depth
    fread(image, sizeof(PIXEL), COMPONENTS, infile);
    fclose(infile);
    return(image);
  } else {
      fputs("Couldn't open an input image or get memory, aborting\n", stderr);
	  exit(1);
  }
}

/* schreiben als ppm datei */

void writeimage(char *filename, PIXEL *image) {
  FILE *outfile = fopen(filename, "w");

  fputs("P6\n", outfile);
  fputs("# CREATOR kludge\n", outfile);
  fputs("400 300\n", outfile);
  fputs("255\n", outfile);
  fwrite(image, sizeof(PIXEL), COMPONENTS, outfile);
  fclose(outfile);
}

/* hauptprogramm */

int main(int argc, char **argv) {
  PIXEL *image1, *image2;
  PIXEL *outimage = (PIXEL*) malloc(IMGSIZE);;

  if (argc == 1) {
      fputs("No command line arguments, aborting!\n", stderr);
      return -1;
  }

  if (!outimage) {
	  fputs("No memory for output image, aborting!\n", stderr);
	  return -1;
  }

  image1 = readimage(argv[1]);
  image2 = readimage(argv[2]);

  if (argv[3][0] == 'b') { // "Uberblenden

      ALPHA alpha = atoi(argv[4]);

      blend(image1, image2, alpha, outimage);

      writeimage(argv[5], outimage);

  } else {  // ChromaKey

      unsigned int key  = atoi(argv[4]);
      unsigned int epsilon  = atoi(argv[5]);

      chromakey(image1, image2, outimage, key, epsilon);

      writeimage(argv[6], outimage);

  }

  free(image1);
  free(image2);
  free(outimage);  
  fputs("Done.\n", stderr);
}


